Require Import Coq.Bool.Bool.
Require Import Coq.Arith.Arith.
Require Import Coq.Lists.List.
Require Import Coq.Program.Equality.

Require Export core.
Require Export labelmap.
Require Export lattice.

Module System.

Import ListNotations.

Import CoreLanguage.
Import LabelMap.
Import SecurityLattice.

Definition ret_addr : Type := option (aid * iid).

Definition get_ret_addr (expect_ret : bool) (ai: aid) (ii: iid) : ret_addr :=
  if expect_ret then Some (ai, ii) else None.


(* An app is a statement *)
Definition app : Type := stm.

(* An app_set is a list of apps *) 
Definition app_set : Type := list app.

(* instance is an app instance that consist of a the instance id, a memory
   a (running) copy of an app, and a return address *)
Inductive instance : Type :=
  | Inst (id: iid) (m: mem) (a: app) (ret: ret_addr).

(* An app_inst_set is a list of running instances of a particular app *)
Definition app_inst_set : Type := list instance.

(* An inst_set is a list of instances running in parallel *)
Definition inst_set : Type := list app_inst_set.

Inductive pendingType : Type := 
| pCall (ra : ret_addr)
| pReturn ( callee_ii : iid).

Inductive pending : Type :=
| Pend (callee : aid) (n : nat) (caller_dl : label) (pt : pendingType).

Definition pqueue : Type := list pending.

Inductive sys : Type :=
  | Sys ( apps: app_set ) (lm: label_map) (insts: inst_set) (queue : pqueue).



Fixpoint not_a_command (s : stm) : Prop :=
  match s with
   | s_cmd (Call _ (e_val _) _ ) | s_cmd (Return (e_val _)) 
     | s_cmd (Exit (e_val _)) | s_cmd (Raise _) | s_cmd (Decl _) =>
      False
   | s_seq s1 s2 => not_a_command s1
   | _ => True
  end.

(* retrieve original statement/code of an app *)
Definition app_stm (apps: list app) (ai : aid): stm :=
  nth ai apps s_null.


(* An instance in an instance list takes a step *)
Inductive app_is_step :
  app_inst_set ->
  app_inst_set -> Prop :=
  | aiss_next inst ais1 ais2 : app_is_step ais1 ais2 -> app_is_step (inst :: ais1) (inst :: ais2)
  | aiss_step ii m1 m2 s1 s2 ret ais:
     not_a_command s1 -> step_stm m1 s1 m2 s2 -> app_is_step ((Inst ii m1 s1 ret) :: ais) ((Inst ii m2 s2 ret) :: ais).

Inductive is_step :
  inst_set ->
  inst_set -> Prop :=
  | iss_next h is1 is2 : is_step is1 is2 -> is_step (h :: is1) (h :: is2)
  | iss_step ais1 ais2 is:  app_is_step ais1 ais2 -> is_step (ais1 :: is) (ais2 :: is).

Definition is_ready (c : cmd) : Prop :=
  match c with
  | Raise _ | Decl _ | Call _ (e_val _) _
  | Return (e_val _) | Exit (e_val _ ) => True
  | _ => False
  end.

(* Decide wether a stament represent an active command 
   and compute the residual command left after executing*)
Inductive is_command:
  stm ->
  cmd -> stm -> Prop :=
  | iscom_c c :
      is_ready c -> is_command (s_cmd c) c s_null
  | iscom_seq s c residual s2:
      is_command s c residual -> is_command (s_seq s s2) c (s_seq residual s2).

(* An instance in an instance set dispatches a command c *)
Inductive app_is_cmd :
  iid ->
  cmd ->
  app_inst_set ->
  app_inst_set -> ret_addr -> Prop :=
  | aisc_next ii c inst ra ais1 ais2 :
      app_is_cmd ii c ais1 ais2 ra ->
      app_is_cmd ii c (inst :: ais1) (inst :: ais2) ra
  | aisc_call ii s callee e er residual m ra ais:
      is_command s (Call callee e er) residual ->
      app_is_cmd ii (Call callee e er) ((Inst ii m s ra) :: ais) ((Inst ii m residual ra) :: ais) ra
  | aisc_raise ii s l residual m ra ais:
      is_command s (Raise l) residual ->
      app_is_cmd ii (Raise l) ((Inst ii m s ra) :: ais) ((Inst ii m residual ra) :: ais) None
  | aisc_decl ii s l residual m ra ais:
      is_command s (Decl l) residual ->
      app_is_cmd ii (Decl l) ((Inst ii m s ra) :: ais) ((Inst ii m residual ra) :: ais) None
  | aisc_return ii s e residual m ra ais:
      is_command s (Return e) residual ->
      app_is_cmd ii (Return e) ((Inst ii m s ra) :: ais) ((Inst ii m s_null ra) :: ais) None
  | aisc_exit ii s e residual m ra ais:
      is_command s (Exit e) residual ->
      app_is_cmd ii (Exit e) ((Inst ii m s ra) :: ais) ((Inst ii nil s_null ra) :: ais) None.

Inductive is_cmd:
 aid -> iid -> cmd -> inst_set -> inst_set -> ret_addr -> Prop :=
 | isc_next ai ii c is1 is2 ra ais:
     is_cmd ai ii c is1 is2 ra ->
     is_cmd (S ai) ii c (ais :: is1) (ais :: is2) ra
 | isc_cmd ii c is ra ais1 ais2:
     app_is_cmd ii c ais1 ais2 ra ->
    is_cmd O ii c (ais1 :: is) (ais2 :: is) ra.

Inductive app_send_intent_existing:
  app_inst_set -> iid -> stm -> nat -> ret_addr -> app_inst_set -> Prop :=
 | asie_next ais1 ii s n ra ais2 inst:
    app_send_intent_existing ais1 ii s n ra ais2 -> app_send_intent_existing (inst::ais1) ii s n ra (inst::ais2)
 | asie_send ais ii s n old_ra ra m:
    app_send_intent_existing ((Inst ii m s_null old_ra) :: ais) ii s n ra
                                     ((Inst ii m (subst_stm ARG (v_nat n) s) ra) ::ais).

Inductive app_send_intent_new:
  app_inst_set -> iid -> stm -> nat -> ret_addr -> app_inst_set -> Prop :=
| asin_send ais ii s n ra:
   app_send_intent_new ais ii s n ra ((Inst ii empty_mem (subst_stm ARG (v_nat n) s) ra) :: (ais)).

Inductive update_app_label_map_and_inst_set:
  label_map -> inst_set ->
  aid -> app_label_map -> app_inst_set ->
  label_map -> inst_set -> Prop :=
  | upd_next lm1 is1 ai new_alm new_ais lm2 is2 alm ais:
      update_app_label_map_and_inst_set lm1 is1 ai new_alm new_ais lm2 is2 ->
      update_app_label_map_and_inst_set
        (alm :: lm1) (ais :: is1) (S ai) new_alm new_ais (alm :: lm2) (ais :: is2)
  | upd_update lm is old_alm new_alm old_ais new_ais:
      update_app_label_map_and_inst_set
        (old_alm :: lm) (old_ais :: is) O new_alm new_ais (new_alm :: lm) (new_ais :: is).

Inductive app_remove_instance:
  list (iid*label) -> app_inst_set -> iid -> list (iid*label) -> app_inst_set -> Prop :=
  | are_next dll1 dll2 ais1 ais2 ii inst_l inst:
(*  beq_iid ii i2 = false -> *)
    app_remove_instance dll1 ais1 ii dll2 ais2 ->
    app_remove_instance ( inst_l :: dll1) (inst :: ais1) ii ( inst_l :: dll2) (inst :: ais2)
| are_remove dll ais ii l m s ra:
    app_remove_instance ( (ii,l) :: dll) ((Inst ii m s ra) :: ais) ii dll ais.

Inductive Busy_ais : iid -> app_inst_set -> Prop :=
  | busya_busy ii m s ra ais:
    s <> s_null -> Busy_ais ii (Inst ii m s ra::ais)
  | busya_next ii ais inst:
    Busy_ais ii ais -> Busy_ais ii (inst::ais).

Inductive Busy : aid -> iid -> inst_set -> Prop :=
  | busy_busy ii is ais:
    Busy_ais ii ais -> Busy O ii (ais::is)
  | busy_next ai ii is ais:
    Busy ai ii is -> Busy (S ai) ii (ais::is).

(* Execute a Call command *)
Inductive execute_call :
  app_set ->
  label_map -> inst_set -> pqueue ->
  label ->
  aid -> nat -> ret_addr ->
  label_map -> inst_set -> pqueue -> Prop :=
  | ecall_none apps lm is pq caller_dl ai n ret:
      ai >= length apps \/ has_app_label_map lm ai NA ->
      execute_call apps lm is pq caller_dl ai n ret lm is pq
  | ecall_sl apps lm is pq caller_dl ai n ra lm2 is2 sl dl ii new_ais:
      has_app_label_map lm ai (Single_Instance_App_Launched sl dl ii) ->
      lte caller_dl dl ->
      app_send_intent_existing (nth ai is []) ii (app_stm apps ai) n ra new_ais ->
      update_app_label_map_and_inst_set lm is ai
          (Single_Instance_App_Launched sl dl ii) new_ais lm2 is2 ->
      execute_call apps lm is pq caller_dl ai n ra lm2 is2 pq
  | ecall_sl_pending apps lm is pq caller_dl ai n ra sl dl ii:
      has_app_label_map lm ai (Single_Instance_App_Launched sl dl ii) ->
      lte caller_dl dl ->
      Busy ai ii is ->
      execute_call apps lm is pq caller_dl ai n ra lm is (Pend ai n caller_dl (pCall ra) :: pq)
  | ecall_sl_block apps lm is pq caller_dl ai n ret sl dl ii:
      has_app_label_map lm ai (Single_Instance_App_Launched sl dl ii) ->
      nlte caller_dl dl ->
      execute_call apps lm is pq caller_dl ai n ret lm is pq
  | ecall_su apps lm is pq caller_dl ai n ra lm2 is2 sl prev new_ais:
      has_app_label_map lm ai (Single_Instance_App_Unlaunched sl prev) ->
      lte caller_dl sl ->
      app_send_intent_new (nth ai is []) (increment_iid prev) (app_stm apps ai) n ra new_ais ->
      update_app_label_map_and_inst_set lm is ai
          (Single_Instance_App_Launched sl sl (increment_iid prev) ) new_ais lm2 is2 ->
      execute_call apps lm is pq caller_dl ai n ra lm2 is2 pq
  | ecall_su_block apps lm is pq caller_dl ai n ret sl prev:
      has_app_label_map lm ai (Single_Instance_App_Unlaunched sl prev) ->
      nlte caller_dl sl ->
      execute_call apps lm is pq caller_dl ai n ret lm is pq
  | ecall_m apps lm is pq caller_dl ai n ra lm2 is2 sl dll ii new_ais:
      has_app_label_map lm ai (Multi_Instance_App sl dll) ->
      lte caller_dl sl ->
      next_iid dll = ii ->
      app_send_intent_new (nth ai is []) ii (app_stm apps ai) n ra new_ais ->
      update_app_label_map_and_inst_set lm is ai
          (Multi_Instance_App sl ( (ii,sl) :: dll) ) new_ais lm2 is2 ->
      execute_call apps lm is pq caller_dl ai n ra lm2 is2 pq
  | ecall_m_block apps lm is pq caller_dl ai n ret sl dll:
      has_app_label_map lm ai (Multi_Instance_App sl dll) ->
      nlte caller_dl sl ->
      execute_call apps lm is pq caller_dl ai n ret lm is pq.

(* Execute a return call *)
Inductive execute_return :
  app_set ->
  label_map -> inst_set -> pqueue ->
  label ->
  aid -> iid -> nat ->
  inst_set -> pqueue -> Prop :=
  | ereturn_sl apps lm is pq caller_dl ai ii n is2 sl dl new_ais:
      has_app_label_map lm ai (Single_Instance_App_Launched sl dl ii) ->
      lte caller_dl dl ->
      app_send_intent_existing (nth ai is []) ii (app_stm apps ai) n None new_ais ->
      update_app_label_map_and_inst_set lm is ai
          (Single_Instance_App_Launched sl dl ii) new_ais lm is2 ->
      execute_return apps lm is pq caller_dl ai ii n is2 pq
  | ereturn_sl_pending apps lm is pq caller_dl ai ii n sl dl:
      has_app_label_map lm ai (Single_Instance_App_Launched sl dl ii) ->
      lte caller_dl dl ->
      Busy ai ii is ->
      execute_return apps lm is pq caller_dl ai ii n is (Pend ai n caller_dl (pReturn ii) :: pq)
  | ereturn_sl_block apps lm is pq caller_dl ai ii n sl dl:
      has_app_label_map lm ai (Single_Instance_App_Launched sl dl ii) ->
      nlte caller_dl dl ->
      execute_return apps lm is pq caller_dl ai ii n is pq
  | ereturn_m apps lm is pq caller_dl ai ii n is2 sl dll dl new_ais:
      has_app_label_map lm ai (Multi_Instance_App sl dll) ->
      has_inst_label dll ii (Some dl) ->
      lte caller_dl dl ->
      app_send_intent_existing (nth ai is []) ii (app_stm apps ai) n None new_ais ->
      update_app_label_map_and_inst_set lm is ai
          (Multi_Instance_App sl dll) new_ais lm is2 ->
      execute_return apps lm is pq caller_dl ai ii n is2 pq
  | ereturn_m_pending apps lm is pq caller_dl ai ii n sl dll dl:
      has_app_label_map lm ai (Multi_Instance_App sl dll) ->
      has_inst_label dll ii (Some dl) ->
      lte caller_dl dl ->
      Busy ai ii is ->
      execute_return apps lm is pq caller_dl ai ii n is (Pend ai n caller_dl (pReturn ii) :: pq)
  | ereturn_m_block apps lm is pq caller_dl ai ii n sl dll dl:
      has_app_label_map lm ai (Multi_Instance_App sl dll) ->
      has_inst_label dll ii (Some dl) ->
      nlte caller_dl dl ->
      execute_return apps lm is pq caller_dl ai ii n is pq.

(* Execute a Raise command *)
Inductive execute_raise :
  label_map ->
  aid -> iid ->
  label ->
  label_map -> Prop :=
  | eraise_s lm ai ii l lm2 sl dl sl2:
      has_app_label_map lm ai (Single_Instance_App_Launched sl dl ii) ->
      lte_bool dl l = true ->
      (nlte l sl /\ sl2 = l \/ lte l sl /\ sl2 = sl) ->
      update_app_label_map lm ai (Single_Instance_App_Launched sl2 l ii) lm2 ->
      execute_raise lm ai ii l lm2
  | eraise_s_block lm ai ii l sl dl:
      has_app_label_map lm ai (Single_Instance_App_Launched sl dl ii) ->
      lte_bool dl l = false ->
      execute_raise lm ai ii l lm
  | eraise_m lm ai ii l lm2 sl dll dl dll2:
     has_app_label_map lm ai (Multi_Instance_App sl dll) ->
     has_inst_label dll ii (Some dl) ->
     lte_bool dl l = true ->
     update_inst_label dll ii l dll2 ->
     update_app_label_map lm ai (Multi_Instance_App sl dll2) lm2 ->
     execute_raise lm ai ii l lm2
  | eraise_m_block lm ai ii l sl dll dl:
     has_app_label_map lm ai (Multi_Instance_App sl dll) ->
     has_inst_label dll ii (Some dl) ->
     lte_bool dl l = false ->
     execute_raise lm ai ii l lm.

Inductive remove_ra_ais:
  app_inst_set -> aid -> iid -> app_inst_set -> Prop :=
  | rrais_nil ai ii: remove_ra_ais [] ai ii []
  | rrais_skip ai ii ai2 ii2 ais1 ais2 ii' m s:
      remove_ra_ais ais1 ai ii ais2 ->
      (beq_aid ai2 ai = false \/ beq_iid ii2 ii = false) ->
      remove_ra_ais (Inst ii' m s (Some (ai2,ii2)) :: ais1) ai ii
                              (Inst ii' m s (Some (ai2,ii2)) :: ais2)
  | rrais_remove ai ii ais1 ais2 ii' m s:
      remove_ra_ais ais1 ai ii ais2 ->
      remove_ra_ais (Inst ii' m s (Some (ai,ii)) :: ais1) ai ii
                              (Inst ii' m s None :: ais2)
  | rrais_none ai ii ais1 ais2 ii' m s:
      remove_ra_ais ais1 ai ii ais2 ->
      remove_ra_ais (Inst ii' m s None :: ais1) ai ii
                              (Inst ii' m s None :: ais2).


Inductive remove_ra_is:
  inst_set -> aid -> iid -> inst_set -> Prop :=
  | rris_nil ai ii: remove_ra_is [] ai ii []
  | rris_con ai ii is1 is2 ais1 ais2:
      remove_ra_ais ais1 ai ii ais2 ->
      remove_ra_is is1 ai ii is2 ->
      remove_ra_is (ais1 :: is1) ai ii (ais2 :: is2).

Inductive remove_pending:
  pqueue -> aid -> iid -> pqueue -> Prop :=
  | rmp_nil ai ii: remove_pending [] ai ii []
  | rmp_call_skip_none ai ii pq1 pq2 ai' n caller_dl:
      remove_pending pq1 ai ii pq2 ->
      remove_pending (Pend ai' n caller_dl (pCall None) :: pq1) ai ii
                                 (Pend ai' n caller_dl (pCall None) :: pq2)
  | rmp_call_skip_some ai ii pq1 pq2 ai' n caller_dl ai2 ii2:
      remove_pending pq1 ai ii pq2 ->
      (beq_aid ai2 ai = false \/ beq_iid ii2 ii = false) ->
      remove_pending (Pend ai' n caller_dl (pCall (Some (ai2,ii2))) :: pq1) ai ii
                                 (Pend ai' n caller_dl (pCall (Some (ai2,ii2))) :: pq2)
  | rmp_call_remove ai ii pq1 pq2 ai' n caller_dl:
      remove_pending pq1 ai ii pq2 ->
      remove_pending (Pend ai' n caller_dl (pCall (Some (ai,ii))) :: pq1) ai ii
                                 (Pend ai' n caller_dl (pCall None) :: pq2)
  | rmp_return_skip ai ii pq1 pq2 ai2 n caller_dl ii2:
      remove_pending pq1 ai ii pq2 ->
      (beq_aid ai2 ai = false \/ beq_iid ii2 ii = false) ->
      remove_pending (Pend ai2 n caller_dl (pReturn ii2) :: pq1) ai ii
                                 (Pend ai2 n caller_dl (pReturn ii2) :: pq2)
  | rmp_return_remove ai ii pq1 pq2 n caller_dl:
      remove_pending pq1 ai ii pq2 ->
      remove_pending (Pend ai n caller_dl (pReturn ii) :: pq1) ai ii pq2.


Inductive execute_exit:
  label_map -> inst_set -> pqueue ->
  aid -> iid ->
  label_map -> inst_set -> pqueue -> Prop :=
  | eexit_s lm lm' ai ii is is' is'' pq pq' sl dl : 
      has_app_label_map lm ai (Single_Instance_App_Launched sl dl ii) ->
      remove_ra_is is ai ii is' ->
      update_app_label_map_and_inst_set lm is' ai
          (Single_Instance_App_Unlaunched sl ii) [] lm' is'' ->
      remove_pending pq ai ii pq' ->
      execute_exit lm is pq ai ii lm' is'' pq'
  | eexit_m lm lm' ai ii is is' is'' pq pq' sl dll dll' new_ais:
      has_app_label_map lm ai (Multi_Instance_App sl dll) ->
      remove_ra_is is ai ii is' ->
      app_remove_instance dll (nth ai is' []) ii dll' new_ais ->
      update_app_label_map_and_inst_set lm is' ai
          (Multi_Instance_App sl dll') new_ais lm' is'' ->
      remove_pending pq ai ii pq' ->
      execute_exit lm is pq ai ii lm' is'' pq'.


Definition out : Type := (nat * label).

(* Execute a comand *)
Inductive execute :
  app_set ->
  label_map -> inst_set -> pqueue ->
  option out ->
  label_map -> inst_set -> pqueue ->
  Prop :=
  | exec_call apps lm1 lm2 is1 is2 pq1 pq2 callee ai ii n expect_ret ra sl dl is':
      is_cmd ai ii (Call callee (e_nat n) expect_ret) is1 is' ra ->
      has_label lm1 ai ii (Some (sl,dl)) ->
      execute_call apps lm1 is' pq1 dl callee n ra lm2 is2 pq2 ->
      execute apps lm1 is1 pq1 (Some (n,dl)) lm2 is2 pq2
  | exec_raise apps lm1 is1 pq ai ii l lm2 is2 ra:
      is_cmd ai ii (Raise l) is1 is2  ra ->
      execute_raise lm1 ai ii l lm2 ->
      execute apps lm1 is1 pq None lm2 is2 pq
  | exec_return apps lm is1 is2 pq1 pq2 ai ii n callee callee_ii sl dl is':
      is_cmd ai ii (Return (e_nat n)) is1 is' (Some (callee,callee_ii)) ->
      has_label lm ai ii (Some (sl,dl)) ->
      execute_return apps lm is' pq1 dl callee callee_ii n is2 pq2 ->
      execute apps lm is1 pq1 (Some (n,dl)) lm is2 pq2
  | exec_return_no_ra apps lm is1 is2 pq ai ii n sl dl:
      is_cmd ai ii (Return (e_nat n)) is1 is2 None ->
      has_label lm ai ii (Some (sl,dl)) ->
      execute apps lm is1 pq (Some (n,dl)) lm is2 pq
  | exec_exit apps lm1 lm2 is1 is2 pq1 pq2 ai ii n callee callee_ii sl dl is' is'' pq':
      is_cmd ai ii (Exit (e_nat n)) is1 is' (Some (callee,callee_ii)) ->
      beq_aid callee ai = false \/ beq_iid callee_ii ii = false ->
      has_label lm1 ai ii (Some (sl,dl)) ->
      execute_exit lm1 is' pq1 ai ii lm2 is'' pq' ->
      execute_return apps lm2 is'' pq' dl callee callee_ii n is2 pq2 ->
      execute apps lm1 is1 pq1 (Some (n,dl)) lm2 is2 pq2
  | exec_exit_no_return apps lm1 lm2 is1 is2 pq1 pq2 ai ii n sl dl is' ra:
      is_cmd ai ii (Exit (e_nat n)) is1 is' ra ->
      ra = (Some (ai,ii)) \/ ra = None ->
      has_label lm1 ai ii (Some (sl,dl)) ->
      execute_exit lm1 is' pq1 ai ii lm2 is2 pq2 ->
      execute apps lm1 is1 pq1 (Some (n,dl)) lm2 is2 pq2.


Inductive process_pending':
  app_set ->
  label_map -> inst_set -> pqueue -> pqueue ->
  label_map -> inst_set -> pqueue -> Prop :=
  | pp_next apps lm1 lm2 is1 is2 pq1 pq2 p pq':
    process_pending' apps lm1 is1 pq1 (pq' ++ [p]) lm2 is2 pq2 ->
    process_pending' apps lm1 is1 (p :: pq1) pq' lm2 is2 pq2
  | pp_call apps lm1 lm2 is1 is2 pq pq' caller_dl callee n ra:
    execute_call apps lm1 is1 (pq' ++ pq) caller_dl callee n ra lm2 is2 (pq' ++ pq) ->
    process_pending' apps lm1 is1 (Pend callee n caller_dl (pCall ra) :: pq) pq'
        lm2 is2 (pq' ++ pq)
  | pp_return apps lm is1 is2 pq pq' callee callee_ii n caller_dl:
    execute_return apps lm is1 (pq' ++ pq) caller_dl callee callee_ii n is2 (pq' ++ pq) ->
    process_pending' apps lm is1 (Pend callee n caller_dl (pReturn callee_ii) :: pq) pq'
        lm is2 (pq' ++ pq).

Definition process_pending apps lm1 is1 pq1 lm2 is2 pq2 :=
  process_pending' apps lm1 is1 pq1 [] lm2 is2 pq2.

Inductive step_type : Type :=
 | tau
 | output (o : out).

(* The whole system takes a step if either an app instance takes an internal step or
    a command is executed *)
Inductive sys_step: sys -> sys -> step_type -> Prop :=
  | ss_tau_iss apps lm is1 is2 pq:
      is_step is1 is2 ->
      sys_step (Sys apps lm is1 pq) (Sys apps lm is2 pq) tau
  | ss_tau_cmd apps lm1 lm2 is1 is2 pq1 pq2:
      execute apps lm1 is1 pq1 None lm2 is2 pq2 ->
      sys_step (Sys apps lm1 is1 pq1) (Sys apps lm2 is2 pq2) tau
  | ss_tau_pend apps lm1 lm2 is1 is2 pq1 pq2 :
      process_pending apps lm1 is1 pq1 lm2 is2 pq2 ->
      sys_step (Sys apps lm1 is1 pq1) (Sys apps lm2 is2 pq2) tau
  | ss_output_cmd apps lm1 lm2 is1 is2 pq1 pq2 v l:
      execute apps lm1 is1 pq1 (Some (v,l)) lm2 is2 pq2 ->
      sys_step (Sys apps lm1 is1 pq1) (Sys apps lm2 is2 pq2) (output (v,l)).

Definition trace := list out.

End System.
